-- new sdram controller now has dm pin available to truncate burst write
-- initialized to fullpage burst read/write.
-- actual burst is limited to 64 32-bit words
-- memory is divided into two regions:
-- a) HTR event buffer
--   each event has 16 512-word HTR buffers
--   there are 128 such buffers
-- absolute address range: 0x??0000 - 0x??7fff in bytes
-- b) monitoring event buffer
--   each event is 8k-word long, a total of 128 buffers are available
-- to access the monitoring data, a PCI memory space of 32k-bytes
-- is assigned. Every event readout starts at relative address of 0x0
-- to fully use the memory available, memory address are calculated
-- with multipliers and adders.
-- absolute address range 0x??8000-0x??ffff in bytes
-- Except for PCI3 accesses, all read/write starts at addresses aligned to
-- 64 word boundary ( six LSBs all zero )
-- VME read cache stores the last 256 32-bit words of VME readout data.
-- It checks the ADDR lines upon rd_ram request, if the data is still
-- in the cache, it takes it and send it out. Cache miss causes retry
-- being sent. It must also monitor the
-- write operation and invalidate the stalemate data.
-- It requests data 16 words at a time.
-- pci3 memory access window is 64kB wide. There are 128 pages.
library IEEE,mylib;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use mylib.vtx_pack.all;
use mylib.dcc_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity sdram_ctl is
    port (
-- reset input, should be connected to the inverted LOCK output of dll
        reset: in STD_LOGIC;
-- random memory test mode
        mem_test: in STD_LOGIC;
        mem_done: in STD_LOGIC;
        mem_fail : out std_logic;
        chk_done : out std_logic;
-- system clock input
        clk: in STD_LOGIC;
        clkn: in STD_LOGIC;
        clk90: in STD_LOGIC;
        clk270: in STD_LOGIC;
-- DLL lock input
        not_locked: in STD_LOGIC;
-- download data readback
        r_data: out STD_LOGIC_VECTOR (31 downto 0);
        rd_debug: in STD_LOGIC;
-- download address
        addr: in STD_LOGIC_VECTOR (13 downto 0);
        n_strobe: in STD_LOGIC;
        flush_cache: in STD_LOGIC;
        rd_ddr_data: in STD_LOGIC;
        cache_chk: in STD_LOGIC;
        cache_hit: out STD_LOGIC;
-- DDR page address
        ddr_pa: in STD_LOGIC_VECTOR (6 downto 0);
-- ddr write queue signals
		  ddr_w_gnt: out std_logic;
		  ddr_wfifo_we: in std_logic;
		  ddr_w_req: in std_logic;
        ddr_w_length: in STD_LOGIC_VECTOR (5 downto 0);
        ddr_w_addr: in STD_LOGIC_VECTOR (20 downto 0);
        ddr_w_data: in STD_LOGIC_VECTOR (31 downto 0);
-- memory readout
        htr_buf_req: in STD_LOGIC;
        htr_buf_gnt: out STD_LOGIC;
        htr_buf_addr: in STD_LOGIC_VECTOR (14 downto 0);
-- htr_buf_length is 32-bit word counts
-- in case of odd 32-bit words, evt_bldr discards the last 32-bit word 
        htr_buf_length: in STD_LOGIC_VECTOR (5 downto 0);
        htr_buf_re: in STD_LOGIC;
        htr_buf_data: out STD_LOGIC_VECTOR (31 downto 0);
-- ram pads
		  ram_pads : inout ram_pad
    );
end sdram_ctl;

architecture sdram_ctl_arch of sdram_ctl is
signal strobe: std_logic;
-- fifos
signal en_wfifo_l: std_logic := '0';
signal wfifo_wa: std_logic_vector(9 downto 0) := (others => '0');
signal wfifo_ra: std_logic_vector(8 downto 0) := (others => '0');
signal wfifo_rap: std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_wa: std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_wap: std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_ra: std_logic_vector(9 downto 0) := (others => '0');
signal ddr_data: std_logic_vector(63 downto 0) := (others => '0');
-- vme cache
signal cache_miss: std_logic := '0';
signal flush_cache_l: std_logic := '0';
signal cache_full: std_logic := '0';
signal ec_cache_bof: std_logic := '0';
-- cache address begin and end, in 32 32-bit word unit
signal d_cache_bof: std_logic_vector(9 downto 0) := (others => '0');
signal pci_addr: std_logic_vector(9 downto 0) := (others => '0');
signal cache_bof: std_logic_vector(13 downto 0) := (others => '0');
signal d_cache_eof: std_logic_vector(13 downto 0) := (others => '0');
signal cache_eof: std_logic_vector(13 downto 0) := (others => '0');
signal cache_ra: std_logic_vector(8 downto 0) := (others => '0');
signal cache_wa: std_logic_vector(8 downto 0) := (others => '0');
signal cache_wap: std_logic_vector(8 downto 0) := (others => '0');
signal cache_data: std_logic_vector(63 downto 0) := (others => '0');
signal vme_rdata: std_logic_vector(31 downto 0) := (others => '0');
signal htr_buf_data_i: std_logic_vector(31 downto 0) := (others => '0');
-- sdram i/o
signal w_addr: std_logic_vector(12 downto 0) := (others => '0');
signal h_addr: std_logic_vector(12 downto 0) := (others => '0');
signal c_addr: std_logic_vector(12 downto 0) := (others => '0');
signal ma: std_logic_vector(12 downto 0) := (others => '0');
signal md_i: std_logic_vector(63 downto 0) := (others => '0');
signal md_iq: std_logic_vector(63 downto 0) := (others => '0');
signal md_o: std_logic_vector(63 downto 0) := (others => '0');
signal dm: std_logic_vector(1 downto 0) := (others => '0');
signal length_0: std_logic := '0';
signal ddr_w_addr_0: std_logic := '0';
signal dqs: std_logic := '0';
signal rasn: std_logic := '0';
signal casn: std_logic := '0';
signal wen: std_logic := '0';
signal csn: std_logic := '0';
signal t_md: std_logic := '0';
signal clke: std_logic := '0';
-- refresh signal generation
signal ref: std_logic := '0';
signal start_init: std_logic := '0';
-- sdram state machine
signal ec_bank: std_logic := '0';
signal is_read: std_logic := '0';
signal is_cache: std_logic := '0';
signal cache_req: std_logic := '0';
signal rqst: std_logic := '0';
signal active: std_logic := '0';
signal active_dl: std_logic := '0';
signal write: std_logic := '0';
signal read: std_logic := '0';
signal en_dll: std_logic := '0';
signal en_dll_dl: std_logic := '0';
signal rst_dll: std_logic := '0';
signal mrs: std_logic := '0';
signal set_ma8: std_logic := '0';
signal set_ma1: std_logic := '0';
signal inh_ras: std_logic := '0';
signal ref_rqst: std_logic := '0';
signal burst_cnt: std_logic_vector(5 downto 0) := (others => '0');
signal burst: std_logic := '0';
signal rd_burst: std_logic := '0';
signal c_burst: std_logic_vector(5 downto 0) := (others => '0');
signal cache_burst: std_logic := '0';
signal cache_burstp: std_logic := '0';
signal h_burst: std_logic := '0';
signal htr_burst: std_logic := '0';
signal htr_burstp: std_logic := '0';
signal wr_burst: std_logic := '0';
signal wr_burstn: std_logic := '0';
signal wr_burstn_dl: std_logic := '0';
signal mode_set: std_logic := '0';
signal pc: std_logic := '0';
signal pc_dl: std_logic := '0';
signal precharge: std_logic := '0';
signal precharge_dl2: std_logic := '0';
signal rfc_cntr: std_logic_vector(3 downto 0) := x"0";
-- random memory test signals
signal lfsr : std_logic_vector(31 downto 0) := x"00000000";
signal ram_data : std_logic_vector(31 downto 0) := x"00000000";
signal tst_addr : std_logic_vector(7 downto 0) := x"00";
signal ec_lfsr: std_logic := '0';
signal chk_done_i: std_logic := '0';
signal mem_test_dl: std_logic := '0';
signal mem_fail_i: std_logic := '0';
signal ec_tst_addr: std_logic := '0';
signal debug1 : std_logic_vector(31 downto 0) := x"00000000";
signal debug2 : std_logic_vector(31 downto 0) := x"00000000";
signal debug3 : std_logic_vector(31 downto 0) := x"00000000";
signal debug4 : std_logic_vector(31 downto 0) := x"00000000";
signal rclk_dcm: std_logic := '0';
signal rclkn_dcm: std_logic := '0';
signal rclk90_dcm: std_logic := '0';
signal rclk: std_logic := '0';
signal rclkn: std_logic := '0';
signal rclk90: std_logic := '0';
signal rclk270: std_logic := '0';
signal dqs_in : std_logic_vector(1 downto 0) := (others => '0');
signal dqs_in_qp : std_logic_vector(1 downto 0) := (others => '0');
signal dqs_in_q : std_logic_vector(1 downto 0) := (others => '0');
signal dqs_status : std_logic_vector(3 downto 0) := (others => '0');
signal ec_dqs: std_logic := '0';
signal psincdec: std_logic := '0';
signal psen: std_logic := '0';
signal psdone: std_logic := '0';
signal ps_rdy: std_logic := '0';
signal ps_cntr : std_logic_vector(7 downto 0) := (others => '0');
signal mem_rdy: std_logic := '0';
signal en_mem_rdy: std_logic := '0';
attribute LOC : string;
attribute LOC of rclkbufg : label is "BUFGMUX0S";
attribute LOC of rclknbufg : label is "BUFGMUX1P";
attribute LOC of rclk90bufg : label is "BUFGMUX3P";
begin
htr_buf_data <= htr_buf_data_i;
mem_fail <= mem_fail_i;
cache_hit <= not cache_miss or mem_test;
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
  end if;
end process;
----------------------------------------------------------------------------------------
-- write logic
ddr_w_gnt <= write;
wfifo_h: ramb16_s36_s36
	  generic map(SIM_COLLISION_CHECK => "NONE")
     port map(
         dia => ddr_w_data, 
         dipa => x"0", 
         addra => wfifo_wa(9 downto 1),
         clka => clk, 
         ssra => '0', 
         ena => wfifo_wa(0), 
         wea => ddr_wfifo_we,
         dib => x"00000000",
         dipb => x"0",
         addrb => wfifo_rap,
         clkb => clk,
         ssrb => '0',
         enb => '1',
         web => '0',
         dob => md_o(63 downto 32));
wfifo_l: ramb16_s36_s36
	  generic map(SIM_COLLISION_CHECK => "NONE")
     port map(
         dia => ddr_w_data, 
         dipa => x"0", 
         addra => wfifo_wa(9 downto 1),
         clka => clk, 
         ssra => '0', 
         ena => en_wfifo_l, 
         wea => ddr_wfifo_we,
         dib => x"00000000",
         dipb => x"0",
         addrb => wfifo_ra,
         clkb => clkn,
         ssrb => '0',
         enb => '1',
         web => '0',
         dob => md_o(31 downto 0));
en_wfifo_l <= not wfifo_wa(0);
process(clk)
begin
  if(clk'event and clk = '1')then
-- wfifo address counters
    if(reset = '1')then
	   wfifo_wa <= (others => '0');
	 elsif(ddr_wfifo_we = '1')then
	   wfifo_wa <= wfifo_wa + 1;
	 end if;
    wfifo_rap <= wfifo_ra;
	 md_iq(31 downto 0) <= md_i(31 downto 0);
  end if;
end process;
process(clkn)
begin
  if(clkn'event and clkn = '1')then
    wr_burstn <= wr_burst;
    if(reset = '1')then
      wfifo_ra <= (others => '0');
    elsif(wr_burstn = '1')then
      wfifo_ra <= wfifo_ra + 1;
    end if;
	 md_iq(63 downto 32) <= md_i(63 downto 32);
  end if;
end process;
--------------------------------------------------------------
rfifo_h: ramb16_s36_s36
	  generic map(SIM_COLLISION_CHECK => "NONE")
     port map(
         dia => x"00000000", 
         dipa => x"0", 
         addra => rfifo_ra(9 downto 1),
         clka => clk, 
         ssra => '0', 
         ena => '1', 
         wea => '0', 
         doa => ddr_data(63 downto 32), 
         dib =>  md_iq(63 downto 32), 
         dipb => x"0",
         addrb => rfifo_wap,
         clkb => clkn,
         ssrb => '0',
         enb => '1',
         web => htr_burstp);
rfifo_l: ramb16_s36_s36
	  generic map(SIM_COLLISION_CHECK => "NONE")
     port map(
         dia => x"00000000", 
         dipa => x"0", 
         addra => rfifo_ra(9 downto 1),
         clka => clk, 
         ssra => '0', 
         ena => '1', 
         wea => '0', 
         doa => ddr_data(31 downto 0), 
         dib =>  md_iq(31 downto 0),
         dipb => x"0",
         addrb => rfifo_wa,
         clkb => clk,
         ssrb => '0',
         enb => '1',
         web => htr_burst);
-- rfifo address counters
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
-- this is to get htr_buf_data out one clock earlier 
		rfifo_ra <= "0000000001";
    elsif(htr_buf_re = '1')then
      rfifo_ra <= rfifo_ra + 1;
    end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
      rfifo_wa <= (others => '0');
    elsif(htr_burst = '1')then
      rfifo_wa <= rfifo_wa + 1;
    end if;
  end if;
end process;
process(clkn)
begin
  if(clkn'event and clkn = '1')then
    rfifo_wap <= rfifo_wa;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
-- please notice that rfifo_ra(0) is actually inverted!!!
    if(rfifo_ra(0) = '0')then
	   htr_buf_data_i <= ddr_data(63 downto 32);
	 else
	   htr_buf_data_i <= ddr_data(31 downto 0);
	 end if;
  end if;
end process;
-- VME read cache
cache_h: ramb16_s36_s36
	  generic map(SIM_COLLISION_CHECK => "NONE")
     port map(
         dia => x"00000000", 
         dipa => x"0", 
         addra => cache_ra,
         clka => clk, 
         ssra => '0', 
         ena => strobe, 
         wea => '0', 
         doa => cache_data(63 downto 32), 
         dib =>  md_iq(63 downto 32), 
         dipb => x"0",
         addrb => cache_wap,
         clkb => clkn,
         ssrb => '0',
         enb => '1',
         web => cache_burstp);
cache_l: ramb16_s36_s36
	  generic map(SIM_COLLISION_CHECK => "NONE")
     port map(
         dia => x"00000000", 
         dipa => x"0", 
         addra => cache_ra,
         clka => clk, 
         ssra => '0', 
         ena => strobe, 
         wea => '0', 
         doa => cache_data(31 downto 0), 
         dib =>  md_iq(31 downto 0),
         dipb => x"0",
         addrb => cache_wa,
         clkb => clk,
         ssrb => '0',
         enb => '1',
         web => cache_burst);
-- cache size is 128 32-bit words
cache_ra <= "00" & addr(7 downto 1) when mem_test = '0' else "00" & tst_addr(7 downto 1);
cache_wa <= "00" & cache_eof(6 downto 0);
i_cache_eof: l_cntr
  generic map(N => 14)
  port map(
    d => d_cache_eof,
    load => flush_cache,
    reset => reset,
    set => '0',
    ec => '1',
    en_cnt => cache_burst,
    clk => clk,
	 q => cache_eof);
d_cache_eof <= '0'& addr(13 downto 5) & "0000";
i_cache_bof: l_cntr
  generic map(N => 10)
  port map(
    d => d_cache_bof,
    load => flush_cache,
    reset => reset,
    set => '0',
    ec => '1',
    en_cnt => ec_cache_bof,
    clk => clk,
	 q => cache_bof(13 downto 4));
cache_bof(3 downto 0) <= "0000";
d_cache_bof <= '0' & addr(13 downto 5);
process(clkn)
begin
  if(clkn'event and clkn = '1')then
    cache_wap <= cache_wa;
  end if;
end process;
i_cache_miss: compWN
    generic map(N => 10)
    port map(
        d => pci_addr,
        u_limit => cache_eof(13 downto 4),
        l_limit => cache_bof(13 downto 4),
        n_good => cache_miss);
pci_addr <= '0' & addr(13 downto 5);
process(clk)
variable case_full :	std_logic_vector(3 downto 0);
begin
  case_full := cache_eof(5 downto 4) & cache_bof(5 downto 4);
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   flush_cache_l <= '1';
	   ec_cache_bof <= '0';
	   cache_full <= '0';
    else
-- flush the cache when cache miss or ddr_pa changed
-- if a cache burst is under way, flush until it ends
	   flush_cache_l <= flush_cache or (flush_cache_l and rqst and is_cache);
		ec_cache_bof <= read and is_cache and cache_full;
	   if(flush_cache = '1')then
	     cache_full <= '0';
		elsif(read = '1' and is_cache = '1')then
		  case case_full is
		    when x"1" | x"6" | x"b" | x"c" =>
			   cache_full <= '1';
			 when others =>
			   null;
		  end case;
		end if;
    end if;
  end if;
end process;
-- vme data
vme_rdata <= cache_data(31 downto 0) when addr(0) = '0' else cache_data(63 downto 32);
r_data <= vme_rdata when rd_ddr_data = '1' else (others => 'Z');
-- ram access request logic
process(clk)
variable case_req : std_logic_vector(3 downto 0);
begin
  case_req := cache_eof(5 downto 4) & addr(6 downto 5);
  if(clk'event and clk = '1')then
    if(reset = '1')then
      htr_buf_gnt <= '0';
      cache_req <= '0';
      rqst <= '0';
      is_read <= '0';
      is_cache <= '0';
    else
      htr_buf_gnt <= active and is_read and not is_cache;
		if(strobe = '1' and cache_chk = '1')then
		  case case_req is
-- request data when empty or near empty
		    when x"0" | x"3" | x"5" | x"4" | x"a" | x"9" | x"f" | x"e" =>
			   cache_req <= '1';
			 when others => 
			   cache_req <= cache_req or mem_test;
		  end case;
	   elsif(flush_cache = '1' or (rqst = '0' and cache_burst = '0'))then
	     cache_req <= '0';
		end if;
      if(precharge = '1' or mode_set = '0')then
        rqst <= '0';
      else
        rqst <= ddr_w_req or htr_buf_req or (cache_req and not cache_burst) or rqst;
      end if;
      if(mode_set = '0')then
		  is_read <= '0';
      elsif(rqst = '0')then
        is_read <= (not ddr_w_req and htr_buf_req) or (cache_req and not cache_burst);
      end if;
      if(rqst = '0')then
        is_cache <= (cache_req and not cache_burst);
      end if;
    end if;
  end if; 
end process;
-- memory address
process(clk)
begin
  if(clk'event and clk = '1')then
    if(active = '1')then
-- always start at even address so that it does not burst backwards
	   w_addr <= ddr_w_addr(20 downto 17) & '0' & ddr_w_addr(7 downto 1) & '0';
	 else
	   w_addr <= ddr_w_addr(20 downto 8);
    end if;
    if(active = '1')then
	   c_addr <= ddr_pa(6 downto 3) & '0' & cache_eof(6 downto 4) & "00000";
	 else
	   c_addr <= ddr_pa(6 downto 0) & cache_eof(12 downto 7);
    end if;
    if(active = '1')then
	   h_addr <= htr_buf_addr(14 downto 11) & '0' & htr_buf_addr(1 downto 0) & "000000";
	 else
	   h_addr <= htr_buf_addr(14 downto 2);
    end if;
  end if;
end process;
process(clk, not_locked)
begin
  if(not_Locked = '1')then
    set_ma1 <= '1';
  elsif(clk'event and clk = '1')then
-- this is used to set to reduced SDRAM driving and read latency
    set_ma1 <= not mrs and set_ma1;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
-- this ensures active, read/write and precharge are to the same bank
    ec_bank <= not rqst;
    if(ec_bank = '1')then
      if(is_read = '0')then
        ma(12) <= w_addr(12);
        ma(11) <= w_addr(11) or en_dll;
      elsif(is_cache = '1')then
	     ma(12 downto 11) <= c_addr(12 downto 11);
	   else
	     ma(12 downto 11) <= h_addr(12 downto 11);
	   end if;
	 end if;
	 en_dll_dl <= en_dll;
	 rst_dll <= en_dll_dl;
    set_ma8 <= en_dll_dl or pc_dl or (rd_burst and not burst);
	 if(set_ma8 = '1')then
	   ma(8) <= '1';
    elsif(is_read = '0')then
      ma(8) <= w_addr(8);
    elsif(is_cache = '1')then
	   ma(8) <= c_addr(8);
	 else
	   ma(8) <= h_addr(8);
	 end if;
    if(is_read = '0')then
      ma(10 downto 9) <= w_addr(10 downto 9);
      ma(7 downto 6) <= w_addr(7 downto 6);
-- latency = 3
      ma(5) <= w_addr(5) or mrs;
      ma(4) <= w_addr(4) or mrs;
-- sequential
      ma(3) <= w_addr(3);
-- full page burst
      ma(2) <= w_addr(2) or mrs;
      ma(1) <= w_addr(1) or set_ma1;
      ma(0) <= w_addr(0) or mrs;
    elsif(is_cache = '1')then
	   ma(10 downto 9) <= c_addr(10 downto 9);
	   ma(7 downto 0) <= c_addr(7 downto 0);
	 else
	   ma(10 downto 9) <= h_addr(10 downto 9);
	   ma(7 downto 0) <= h_addr(7 downto 0);
	 end if;
  end if;
end process;
-- generate read and write burst signals -------------------------------------------------------
-- burst_cntr
burst <= burst_cnt(5);
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   burst_cnt <= "000000";
	 elsif(active = '1')then	 
	   if(is_cache = '1')then
		  burst_cnt <= "110000";
		elsif(is_read = '0')then
		  burst_cnt <= '1' & not ddr_w_length(5 downto 1);
		else
		  burst_cnt <= '1' & not htr_buf_length(5 downto 1);
		end if;
	 elsif(burst = '1')then
      burst_cnt <= burst_cnt + 1;	 
    end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
      active <= '0';
      active_dl <= '0';
      rd_burst <= '0';
      h_burst <= '0';
      c_burst <= "000000";
	   cache_burst <= '0';
      wr_burst <= '0';
	 else
      active <= not rasn and casn and wen;
      active_dl <= active;
      rd_burst <= burst and is_read;
      h_burst <= burst and is_read and not is_cache;
      c_burst(0) <= burst and is_cache and not flush_cache_l;
		c_burst(1) <= c_burst(0) and not flush_cache_l;
		c_burst(2) <= c_burst(1) and not flush_cache_l;
		c_burst(3) <= c_burst(2) and not flush_cache_l;
		c_burst(4) <= c_burst(3) and not flush_cache_l;
		c_burst(5) <= c_burst(4) and not flush_cache_l;
      cache_burst <= not flush_cache and not flush_cache_l and c_burst(5);
      wr_burst <= burst and not is_read;
    end if;
  end if;
end process;
i_htr_burst: delayN
  generic map(N=> 6) port map(d => h_burst, ce => '1', c => clk, q => htr_burst);
process(clkn)
begin
  if(clkn'event and clkn = '1')then
    htr_burstp <= htr_burst;
    cache_burstp <= cache_burst;
  end if;
end process;
-- sdram state machine -----------------------------------------------------------------------------------------
-- sdram power on initialization states: precharge all banks, enable DLL, reset DLL, precharge all banks,
-- two refresh cycle and set mode register.
i_en_dll: delayN
  generic map(N=> 7) port map(d => start_init, ce => '1', c => clk, q => en_dll);
i_precharge_dl2: delayN
  generic map(N=> 2) port map(d => precharge, ce => '1', c => clk, q => precharge_dl2);
process(clk,not_locked)
variable state: std_logic_vector(2 downto 0);
begin
state := rasn & casn & wen;
  if(not_locked = '1')then
    write <= '0';
	 read <= '0';
    pc <= '0';
    pc_dl <= '0';
    precharge <= '0';
    rfc_cntr <= "0000";
	 inh_ras <= '1';
    rasn <= '1';
    casn <= '1';
    wen <= '1';
    mode_set <= '0';
    ref_rqst <= '0';
  elsif(clk'event and clk = '1')then
	 write <= rasn and not casn and not wen;
	 read <= rasn and not casn and wen;
    pc <= (wr_burst and not burst) or start_init or rst_dll;
	 pc_dl <= pc;
    precharge <= pc_dl or (rd_burst and not burst);
-- auto refresh tRFC counter
	 rfc_cntr(3) <= (not rasn and not casn) or (rfc_cntr(3) and not (rfc_cntr(2) and rfc_cntr(1)));
    rfc_cntr(2) <= (rfc_cntr(2) xor (rfc_cntr(1) and rfc_cntr(0))) and rfc_cntr(3);
    rfc_cntr(1) <= (rfc_cntr(1) xor rfc_cntr(0)) and rfc_cntr(3);
    rfc_cntr(0) <= not rfc_cntr(0) and rfc_cntr(3);
-- inhibit RAS after refresh or read/write until tRFC or tPR
    inh_ras <= not rasn or (inh_ras and not precharge_dl2 and not (rfc_cntr(2) and rfc_cntr(1))) ;
    if(mrs = '1')then
      mode_set <= '1';
    end if;
-- latch refresh request
    if(ref = '1')then
      ref_rqst <= '1';
    end if;
-- default to go to NOP
    rasn <= '1';
    casn <= '1';
    wen <= '1';
    case state is
-- NOP
      when "111" =>
---- goto precharge
        if(precharge = '1')then
          rasn <= '0';
          wen <= '0';
        end if;
---- goto auto refresh
        if(ref_rqst = '1' and inh_ras = '0')then
          rasn <= '0';
          casn <= '0';
        end if;
---- goto set mode
        if(en_dll = '1' or rst_dll = '1' or mrs = '1')then
          rasn <= '0';
          casn <= '0';
          wen <= '0';
        end if;
---- goto active
        if(rqst = '1' and ref_rqst = '0' and inh_ras = '0')then
          rasn <= '0';
        end if;
---- goto write or read
        if(active_dl = '1')then
          casn <= '0';
          wen <= is_read;
        end if;
-- READ & WRITE
      when "101" | "100" =>
-- ACTIVE
      when "011" =>
-- AUTO REFRESH
      when "001" =>
			ref_rqst <= '0';
-- PRECHARGE
-- LOAD MODE REGISTER
      when others =>
      null;
    end case;
  end if;
end process;
----------------------------------------------------------------------------------------
-- ram pads
ram_pads.csn <= csn;
ram_pads.clke <= clke;
process(clk)
begin
  if(clk'event and clk = '1')then
    t_md <= (not wr_burst and (t_md or not rasn)) or not mode_set;
    ddr_w_addr_0 <= (not active and ddr_w_addr_0) or (active and ddr_w_addr(0)); 
    length_0 <= (not active and length_0) or (active and ddr_w_length(0)); 
-- if odd number of words(i.e. length(0)='0'), mask off the last word
	 dm(1) <= not wr_burstn_dl or (not wr_burstn and not length_0 and not ddr_w_addr_0);
    dqs <= not t_md;
  end if;
end process;
process(clkn)
begin
  if(clkn'event and clkn = '1')then
    dm(0) <= (not wr_burstn and not ddr_w_addr_0) or ((not wr_burstn_dl or (not length_0 and not wr_burstn)) and ddr_w_addr_0);
    wr_burstn_dl <= wr_burstn;
  end if;
end process;
i_dm: ddr_io
    port map(
          clk => clkn,
          clkn => clk,
          pad => ram_pads.dm,
			 t => '0',
          d0 => dm(0),
          d1 => dm(1),
          iclk => '0',
          iclkn => '0');
i_dqs: ddr_io
    port map(
          clk => clk270,
          clkn => clk90,
          pad => ram_pads.dqs,
			 t => t_md,
          d0 => dqs,
          d1 => '0',
          iclk => rclk270,
          iclkn => rclk90,
          q0 => dqs_in(0),
          q1 => dqs_in(1));
process(clk)
begin
  if(clk'event and clk = '1')then
    ram_pads.wen <= wen;
    ram_pads.casn <= casn;
    ram_pads.rasn <= rasn;
    ram_pads.addr <= ma;
  end if;
end process;
g_md: for i in 0 to 31 generate
  ddr_data: ddr_io
    port map(
          clk => clk,
          clkn => clkn,
          pad => ram_pads.data(i),
			 t => t_md,
          d0 => md_o(32+i),
          d1 => md_o(i),
          iclk => rclk,
          iclkn => rclkn,
          q0 => md_i(i),
          q1 => md_i(32+i));
end generate g_md;
----------------------------------------------------------------------------------------
-- instantiate refresh generator
I_refresh: refresh PORT MAP(
		clk => clk,
		not_locked => not_locked,
		mode_set => mode_set,
		mrs => mrs,
		start_init => start_init,
		ddr_csn => csn,
		ddr_clke => clke,
		ref => ref
	);
process(rclk270)
begin
  if(rclk270'event and rclk270 = '1')then
    dqs_in_qp <= dqs_in;
  end if;
end process;
process(rclkn)
begin
  if(rclkn'event and rclkn = '1')then
    dqs_in_q <= dqs_in_qp;
  end if;
end process;
i_ec_dqs: delayN
  generic map(N=> 7) port map(d => rd_burst, ce => '1', c => clk, q => ec_dqs);
process(clkn,not_locked)
begin
  if(not_locked = '1')then
    mem_rdy <= '0';
    en_mem_rdy <= '0';
    ps_rdy <= '1';
    ps_cntr <= x"a8";
	 dqs_status <= x"0";
	 psen <= '0';
  elsif(clkn'event and clkn = '1')then
    ps_rdy <= not psen and (ps_rdy or psdone);
	 if(psen = '1')then
	   if(psincdec = '1')then
	     ps_cntr <= ps_cntr + 1;
	   else
	     ps_cntr <= ps_cntr - 1;
	   end if;
	 end if;
	 if(psdone = '1' or ref = '1')then
	   dqs_status <= x"0";
	 elsif(ec_dqs = '1')then
	   dqs_status(0) <= dqs_status(0) or not dqs_in_q(0);
	   dqs_status(1) <= dqs_status(1) or dqs_in_q(0);
	   dqs_status(2) <= dqs_status(2) or not dqs_in_q(1);
	   dqs_status(3) <= dqs_status(3) or dqs_in_q(1);
		en_mem_rdy <= '1';
	 end if;
	 if(ref = '1' and ps_rdy = '1')then
	   case dqs_status is
		  when "0000" => null;	 
		  when "1001" =>
		    psen <= '1';
			 psincdec <= '1';
			 if(en_mem_rdy = '1' and psincdec = '0')then
			   mem_rdy <= '1';
			 end if;
		  when "0110" =>
		    psen <= '1';
			 psincdec <= '0';
			 if(en_mem_rdy = '1' and psincdec = '1')then
			   mem_rdy <= '1';
			 end if;
		  when others => mem_rdy <= '1';	 
		end case;
	 else
	   psen <= '0';
	 end if;
  end if;
end process;
rclk270 <= not rclk90;
rclkbufg : BUFG port map (O=>rclk, I=>rclk_dcm);
rclknbufg : BUFG port map (O=>rclkn, I=>rclkn_dcm);
rclk90bufg : BUFG port map (O=>rclk90, I=>rclk90_dcm);
DCMrclk: dcm 
   generic map(CLKOUT_PHASE_SHIFT => "VARIABLE",
					PHASE_SHIFT => 40,
				   STARTUP_WAIT => TRUE)  
	port map	
         ( CLKIN   => clk90 ,
           CLKFB   => rclk ,
           RST     => '0' ,
           DSSEN   => '0' ,
           PSINCDEC=> psincdec ,
           PSEN    => psen ,
           PSCLK   => clkn ,
			  clk0    => rclk_dcm,
           CLK90   => rclk90_dcm,
           CLK180   => rclkn_dcm,
			  PSDONE  => psdone);
-- randome memory test
i_ec_lfsr: delayN generic map( N => 4)    port map( d => ec_tst_addr, ce => '1', c => clk, q => ec_lfsr);
i_chk_done: delayN generic map( N => 6)    port map( d => chk_done_i, ce => '1', c => clk, q => chk_done);
process(clk)
begin
  if(clk'event and clk = '1')then
    if(mem_test = '0')then
	   tst_addr <= (others => '0');
	 elsif(ec_tst_addr = '1')then
	   tst_addr <= tst_addr + 1;
	 end if;
	 if(strobe = '1')then
	   mem_test_dl <= mem_test;
	 end if;
	 ec_tst_addr <= mem_test_dl and not strobe and (cache_eof(4) xor tst_addr(5));
    if(mem_test = '0')then
	   ram_data <= (others => '0');
	   chk_done_i <= '0';
	 elsif(strobe = '1')then
	   if(tst_addr(0) = '0')then
	     ram_data <= cache_data(63 downto 32);
	   else
	     ram_data <= cache_data(31 downto 0);
		end if;
	   if(tst_addr(4 downto 0) = "11111")then
	     chk_done_i <= '1';
		else
	     chk_done_i <= '0';
		end if;
	 end if;
	 if(reset = '1')then
	   mem_fail_i <= '0';
	 elsif(ec_lfsr = '1' and lfsr /= ram_data)then
		  mem_fail_i <= '1';
	 end if;
	 if(mem_test = '0')then
	   lfsr <= (others => '0');
	 elsif(ec_lfsr = '1')then
	   lfsr <= lfsr(30 downto 0) & not(lfsr(31) xor lfsr(21) xor lfsr(1) xor lfsr(0));
	 end if;
--	 if(reset = '1')then
--	   debug1 <= x"00000000";
--	 elsif(ddr_wfifo_we = '1' and ddr_w_data(31 downto 0) = x"a00002d1")then
--	   debug1 <= debug1 + 1;
--	 end if;	
--	 if(reset = '1')then
--	   debug2 <= x"00000000";
--	 elsif(htr_buf_re = '1' and htr_buf_data_i = x"a00002d1")then
--	   debug2 <= debug2 + 1;
--	 end if;	
	 if(ec_lfsr = '1' and mem_fail_i = '0')then
	   debug1 <= ram_data;
		debug2 <= lfsr;
		debug3(31 downto 16) <= tst_addr & '0' & ddr_pa;
		debug3(13 downto 0) <= cache_eof;
		debug4 <= debug1;
	 end if;
	 if(strobe = '1')then
	   debug3(15 downto 14) <= mem_fail_i & mem_done;
	 end if;
  end if;
end process;

r_data(31 downto 16) <= "000000" & mem_rdy & ps_rdy & ps_cntr when rd_debug = '1' and addr(2 downto 0) = "011" else 
								"ZZZZZZZZZZZZZZZZ";
r_data <= debug1 when rd_debug = '1' and addr(2 downto 0) = "100" else (others => 'Z');
r_data <= debug2 when rd_debug = '1' and addr(2 downto 0) = "101" else (others => 'Z');
r_data <= debug3 when rd_debug = '1' and addr(2 downto 0) = "110" else (others => 'Z');
r_data <= debug4 when rd_debug = '1' and addr(2 downto 0) = "111" else (others => 'Z');
end sdram_ctl_arch;
